"""
Synthesis methods.

Classes to perform stimulus synthesis.
"""

# ignore unused imports
# ruff: noqa: F401
from .eigendistortion import Eigendistortion
from .mad_competition import MADCompetition
from .metamer import Metamer, MetamerCTF
